/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozCmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "CommandDataSrc.h"
#include "ComputedFieldZincXpidl.hpp"
#include "NodeSrc.h"
#include "ElementZincXpidl.hpp"
#include "RegionSrc.h"
#include "SceneViewerSrc.h"
#include "TimeSequenceZincXpidl.hpp"
#include "zincCmguiSrc.h"
#include "nsMemory.h"
#include <stdlib.h>
#ifndef WIN32
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gtk/gtk_cmiss_scene_viewer.h>
#endif
extern "C" {
#include <command/cmiss.h>
#include <computed_field/computed_field.h>
#include <computed_field/computed_field_window_projection.h>
#include <selection/node_selection.h>
}

class CmguiCommandDataData
{
public:
	int scene_viewer_count;

	CmguiCommandDataData() 
	{
		scene_viewer_count = 0;
	}
};

CmguiCommandData::CmguiCommandData()
{

	char* cmgui_argv[] = {strdup("cmgui"), strdup("-server"), NULL};

	// Developers may prefer to have a command window which will happen
	// if -server is not specified i.e. use the following line.
	//	char* cmgui_argv[] = {strdup("cmgui"), NULL};

	char **string_ptr;

#ifdef WIN32
	impl = create_Cmiss_command_data((sizeof(cmgui_argv)/sizeof(char *) - 1),
		cmgui_argv, "0.0", GetModuleHandle("libcmgui.dll"), NULL, NULL, 0);
#else
	impl = create_Cmiss_command_data((sizeof(cmgui_argv)/sizeof(char *) - 1),
		cmgui_argv, "0.0"); 
#endif

	// Free the memory from the strdup's
	string_ptr = cmgui_argv;
	while (*string_ptr)
	{
		free(*string_ptr);
		string_ptr++;
	}

	data = new CmguiCommandDataData();
}

CmguiCommandData::~CmguiCommandData()
{
	if (data)
	{
		delete data;
	}
	if (impl)
	{
		DESTROY(Cmiss_command_data)(&impl);
	}
}

/* void defineMemoryBlock (in string name, [array, size_is (sizeof_data)] in octet data, in PRUint32 sizeof_data); */
NS_IMETHODIMP CmguiCommandData::DefineMemoryBlock(const char *name, PRUint8 *data, PRUint32 sizeof_data)
{
    struct IO_stream_package* package;
    package = Cmiss_command_data_get_IO_stream_package(impl);
    if (package == NULL)
		 return NS_ERROR_UNEXPECTED;
      
    if (!IO_stream_package_define_memory_block(package, (char*)name,
                                               (void *)data, sizeof_data))
	 {
		 return NS_ERROR_UNEXPECTED;
	 }
	 else
	 {
		 return NS_OK;
	 }
}

/* [cmguihandcoded] void freeMemoryBlock (in string name); */
NS_IMETHODIMP CmguiCommandData::FreeMemoryBlock(const char *name)
{
    struct IO_stream_package* package;
    if (!(package = Cmiss_command_data_get_IO_stream_package(impl)))
	 {
		 return NS_ERROR_UNEXPECTED;
	 }

    if (!IO_stream_package_free_memory_block(package, (char*)name))
	 {
		 return NS_ERROR_UNEXPECTED;
	 }

	 return NS_OK;
}

/* CmguiITimeSequencePackage createTimeSequencePackage (); */
NS_IMETHODIMP CmguiCommandData::CreateTimeSequencePackage(CmguiITimeSequencePackage **_retval)
{
	Cmiss_time_sequence_package_id tsp
      = CREATE(Cmiss_time_sequence_package)();
	if (!(*_retval = new CmguiTimeSequencePackage(tsp)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	NS_ADDREF(*_retval);
	return NS_OK;
}

/* void executeCommand (in string command); */
NS_IMETHODIMP CmguiCommandData::ExecuteCommand(const char *command)
{
	struct Execute_command *ec;
    ec = Cmiss_command_data_get_execute_command(impl);
    if (ec == NULL)
      return NS_ERROR_UNEXPECTED;

    if (!Execute_command_execute_string(ec, (char*)command))
      return NS_ERROR_UNEXPECTED;

	return NS_OK;
}

/* [cmguihandcoded] void createSceneViewer (in zincIPlugin plugin, in CmguiISceneViewerResultMonitor svrm); */
NS_IMETHODIMP CmguiCommandData::CreateSceneViewer(zincIPlugin *plugin, 
	CmguiISceneViewerResultMonitor *svrm)
{
    if (!svrm)
    {
      return NS_ERROR_UNEXPECTED;
    }
    CmguiSceneViewer* svi;
	 if ((svi = createSceneViewerFromPlugin( 
				Cmiss_command_data_get_scene_viewer_package(impl), plugin,
				this)))
	 {
		 NS_ADDREF(svi);
		 svrm->SceneViewerReady(svi);
	 }
	 else
	 {
		 svrm->SceneViewerFailed();
	 }

    return NS_OK;
}



/* CmguiIComputedField findFieldByName (in string name); */
NS_IMETHODIMP CmguiCommandData::FindFieldByName(const char *name, CmguiIComputedField **_retval)
{
	struct Computed_field* cf;
	struct MANAGER(Computed_field)* cfm;
	
	cfm = Cmiss_command_data_get_computed_field_manager(impl);
	if ( (cf = Cmiss_computed_field_manager_get_field(cfm, name)) ) {

		if (!(*_retval = new CmguiComputedField(cf))) {
			return NS_ERROR_UNEXPECTED;
		} else {
			NS_ADDREF(*_retval);
		}

	}	else {
		// try setting return value to null if field not found
		*_retval = nsnull;		

	}
	
	return NS_OK;
}

/* CmguiIComputedField findFieldByName (in string name); */
NS_IMETHODIMP CmguiCommandData::IsFieldDefined(const char *name, PRBool *_retval)
{
	struct Computed_field* cf;
	struct MANAGER(Computed_field)* cfm;
	
	cfm = Cmiss_command_data_get_computed_field_manager(impl);
	if ( (cf = Cmiss_computed_field_manager_get_field(cfm, name)) ) {

		*_retval = true;
	}
	else {
		*_retval = false;		
	}

	return NS_OK;
}



/* CmguiIComputedField createField (in string name); */
NS_IMETHODIMP CmguiCommandData::CreateField(const char *name, CmguiIComputedField **_retval)
{
	struct Computed_field* cf;
	struct MANAGER(Computed_field)* cfm;
	
	cfm = Cmiss_command_data_get_computed_field_manager(impl);
	cf = CREATE(Computed_field)(name);
	
	if (!(*_retval = new CmguiComputedField(cf))) {
		
		return NS_ERROR_UNEXPECTED;
		
	} else {
	
		ADD_OBJECT_TO_MANAGER(Computed_field)(cf, cfm);		
		NS_ADDREF(*_retval);
		return NS_OK;
	}

}

/* [cmguihandcoded] CmguiIComputedField createWindowProjectionField (in string name, in CmguiIComputedField sourceField, in CmguiISceneViewer sViewer, in unsigned long projType); */
NS_IMETHODIMP CmguiCommandData::CreateWindowProjectionField(const char *name, CmguiIComputedField *sourceField, CmguiISceneViewer *sViewer, PRUint32 projType, CmguiIComputedField **_retval)
{
  CmguiComputedField *sourceField_cast = NULL;
  sourceField->QueryInterface(NS_GET_IID(CmguiComputedField), (void **)&sourceField_cast);
  if (NULL == sourceField_cast)
  {
     return NS_ERROR_UNEXPECTED;
  }
  CmguiSceneViewer *sViewer_cast = NULL;
  sViewer->QueryInterface(NS_GET_IID(CmguiSceneViewer), (void **)&sViewer_cast);
  if (NULL == sViewer_cast)
  {
     return NS_ERROR_UNEXPECTED;
  }

  // This is not ACCESSd, so don't need to DEACCESS.
  struct MANAGER(Computed_field)* cfm = Cmiss_command_data_get_computed_field_manager(impl);

  // Actually create the computed field...
  // Yet another char* that should have been const char*...
  struct Computed_field* cf = ACCESS(Computed_field)(CREATE(Computed_field)(const_cast<char*>(name)));
  if (!cf)
  {
    return NS_ERROR_UNEXPECTED;
  }
  
  // Set it to be a window projection field...
  if (!(Computed_field_set_type_window_projection(
	  cf, sourceField_cast->impl, sViewer_cast->impl, "none_ZINC_sceneviewer",
     0, (Computed_field_window_projection_type)projType, cfm)))
  {
    return NS_ERROR_UNEXPECTED;
  }

  // Add it to the manager...
  ADD_OBJECT_TO_MANAGER(Computed_field)(cf, cfm);

  *_retval = new CmguiComputedField(cf);
	NS_ADDREF(*_retval);
  return NS_OK;
}

static void process_node_selection(FE_node_selection* nodeSelection,
	FE_node_selection_changes* changes, void* userData)
{
  // userData is an event handler...
	CmguiIEventHandler *event_handler =
    static_cast<CmguiIEventHandler*>(userData);

  event_handler->OnEvent();
}

static void process_element_selection(FE_element_selection* elementSelection,
	FE_element_selection_changes* changes, void* userData)
{
  // userData is an event handler...
	CmguiIEventHandler *event_handler =
    static_cast<CmguiIEventHandler*>(userData);

  event_handler->OnEvent();
}

/* [cmguihandcoded] void registerSelectionHandler (in CmguiIEventHandler handler); */
NS_IMETHODIMP CmguiCommandData::RegisterSelectionHandler(CmguiIEventHandler *handler)
{
	FE_node_selection* fns = Cmiss_command_data_get_node_selection(impl);
	if (!fns)
	{
      return NS_ERROR_UNEXPECTED;
	}

	if (!FE_node_selection_add_callback(fns, process_node_selection,
			static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;
	}

	FE_element_selection* fes = Cmiss_command_data_get_element_selection(impl);
	if (!fes)
	{
      return NS_ERROR_UNEXPECTED;
	}
	if (!FE_element_selection_add_callback(fes, process_element_selection,
			static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;
	}

	NS_ADDREF(handler);
	return NS_OK;
}

/* [cmguihandcoded] void deregisterSelectionHandler (in CmguiIEventHandler handler); */
NS_IMETHODIMP CmguiCommandData::DeregisterSelectionHandler(CmguiIEventHandler *handler)
{
	FE_node_selection* fns = Cmiss_command_data_get_node_selection(impl);
	if (!fns)
	{
      return NS_ERROR_UNEXPECTED;
	}
	if (!FE_node_selection_remove_callback(fns, process_node_selection,
			static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;		
	}

	FE_element_selection* fes = Cmiss_command_data_get_element_selection(impl);
	if (!fes)
	{
      return NS_ERROR_UNEXPECTED;
	}
	if (!FE_element_selection_remove_callback(fes, process_element_selection,
			static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;		
	}

	NS_RELEASE(handler);
	return NS_OK;
}

struct node_list_iteration_state
{
  CmguiINode **node_array;
  PRUint32 nextIndex;
};

static int node_list_iteration(Cmiss_node_id node, void* userData)
{
  node_list_iteration_state* nlis =
    static_cast<node_list_iteration_state*>(userData);

  nlis->node_array[nlis->nextIndex] = new CmguiNode(ACCESS(FE_node)(node));
  NS_ADDREF(nlis->node_array[nlis->nextIndex]);
  nlis->nextIndex++;
  return 1;
}

/* [cmguihandcoded] void getSelectedNodes ([array, size_is (sizeof_selectedNodes)] out CmguiINode selectedNodes, out PRUint32 sizeof_selectedNodes); */
NS_IMETHODIMP CmguiCommandData::GetSelectedNodes(CmguiINode ***selectedNodes, PRUint32 *sizeof_selectedNodes)
{
	// Firstly, get the data structure out of cmgui...
	FE_node_selection* fns = Cmiss_command_data_get_node_selection(impl);
	if (fns == NULL)
	{
      return NS_ERROR_UNEXPECTED;
	}

	struct LIST(FE_node)* fnl = FE_node_selection_get_node_list(fns);
	if (fnl == NULL)
	{
      return NS_ERROR_UNEXPECTED;
	}

	struct node_list_iteration_state nlis;
	int length = NUMBER_IN_LIST(FE_node)(fnl);

	nlis.node_array = (CmguiINode **) nsMemory::Alloc(sizeof (CmguiINode *) * length);
	nlis.nextIndex = 0;
	if (!(FOR_EACH_OBJECT_IN_LIST(FE_node)(node_list_iteration, &nlis, fnl)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	*selectedNodes = nlis.node_array;
	*sizeof_selectedNodes = length;
	return NS_OK;
}

struct element_list_iteration_state
{
  CmguiIElement **element_array;
  PRUint32 nextIndex;
};

static int element_list_iteration(Cmiss_element_id element, void* userData)
{
  element_list_iteration_state* nlis =
    static_cast<element_list_iteration_state*>(userData);

  nlis->element_array[nlis->nextIndex] = new CmguiElement(ACCESS(FE_element)(element));
  NS_ADDREF(nlis->element_array[nlis->nextIndex]);
  nlis->nextIndex++;
  return 1;
}

/* [cmguihandcoded] void getSelectedElements ([array, size_is (sizeof_selectedElements)] out CmguiIElement selectedElements, out PRUint32 sizeof_selectedElements); */
NS_IMETHODIMP CmguiCommandData::GetSelectedElements(CmguiIElement ***selectedElements, PRUint32 *sizeof_selectedElements)
{
	// Firstly, get the data structure out of cmgui...
	FE_element_selection* fes = Cmiss_command_data_get_element_selection(impl);
	if (fes == NULL)
	{
      return NS_ERROR_UNEXPECTED;
	}

	struct LIST(FE_element)* fnl = FE_element_selection_get_element_list(fes);
	if (fnl == NULL)
	{
      return NS_ERROR_UNEXPECTED;
	}

	struct element_list_iteration_state nlis;
	int length = NUMBER_IN_LIST(FE_element)(fnl);

	nlis.element_array = (CmguiIElement **) nsMemory::Alloc(sizeof (CmguiIElement *) * length);
	nlis.nextIndex = 0;
	if (!(FOR_EACH_OBJECT_IN_LIST(FE_element)(element_list_iteration, &nlis, fnl)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	*selectedElements = nlis.element_array;
	*sizeof_selectedElements = length;
	return NS_OK;
}

void CmguiCommandData_opening_scene_viewer_notification(CmguiCommandData *parent, 
	CmguiSceneViewer *cmguiSceneViewer)
{
	parent->data->scene_viewer_count++;
}

void CmguiCommandData_closing_scene_viewer_notification(CmguiCommandData *parent, 
	CmguiSceneViewer *cmguiSceneViewer)
{
	parent->data->scene_viewer_count--;
	if (parent->data->scene_viewer_count == 0)
	{
		/* We have to destroy the command data now while the graphics context is
			still current */
		DESTROY(Cmiss_command_data)(&(parent->impl));
	}
}
